import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.Socket;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Base64;
import java.util.Hashtable;

import javax.xml.bind.DatatypeConverter;

import org.apache.commons.io.input.CountingInputStream;
import org.apache.commons.io.output.CountingOutputStream;
import org.bouncycastle.asn1.ASN1Encodable;
import org.bouncycastle.asn1.ASN1ObjectIdentifier;
import org.bouncycastle.asn1.ASN1OctetString;
import org.bouncycastle.asn1.ocsp.BasicOCSPResponse;
import org.bouncycastle.asn1.ocsp.SingleResponse;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.Extensions;
import org.bouncycastle.asn1.x509.KeyPurposeId;
import org.bouncycastle.crypto.tls.Certificate;
import org.bouncycastle.crypto.tls.CertificateStatusRequest;
import org.bouncycastle.crypto.tls.CertificateStatusType;
import org.bouncycastle.crypto.tls.DefaultTlsClient;
import org.bouncycastle.crypto.tls.ExtensionType;
import org.bouncycastle.crypto.tls.OCSPStatusRequest;
import org.bouncycastle.crypto.tls.ServerOnlyTlsAuthentication;
import org.bouncycastle.crypto.tls.TlsAuthentication;
import org.bouncycastle.crypto.tls.TlsClient;
import org.bouncycastle.crypto.tls.TlsClientProtocol;
import org.bouncycastle.crypto.tls.TlsExtensionsUtils;

public class TLSClient {
	private static final ASN1ObjectIdentifier GOOGLE = new ASN1ObjectIdentifier("1.3.6.1.4.1.11129");
	public static final ASN1ObjectIdentifier X509_SIGNED_CERTIFICATE_TIMESTAMP = GOOGLE.branch("2.4.2");
	public static final ASN1ObjectIdentifier PRECERTIFICATE_POISON = GOOGLE.branch("2.4.3");
	public static final KeyPurposeId PRECERTIFICATE_SIGNING_CERTIFICATE = KeyPurposeId.getInstance(GOOGLE.branch("2.4.4"));
	public static final ASN1ObjectIdentifier OCSP_SIGNED_CERTIFICATE_TIMESTAMP = GOOGLE.branch("2.4.5");

	private static Extension getExtension(Extensions extensions, ASN1ObjectIdentifier oid) {
		if (extensions == null) {
			return null;
		}

		return extensions.getExtension(oid);
	}

	private String host;
	private boolean usedOCSPStapling = false;
	/**
	 * @return the host
	 */
	public String getHost() {
		return host;
	}

	/**
	 * @return the usedOCSPStapling
	 */
	public boolean usedOCSPStapling() {
		return usedOCSPStapling;
	}

	/**
	 * @return the usedTLSExtension
	 */
	public boolean usedTLSExtension() {
		return usedTLSExtension;
	}

	/**
	 * @return the usedX509v3Extension
	 */
	public boolean usedX509v3Extension() {
		return usedX509v3Extension;
	}

	/**
	 * @return the oCSPSCT
	 */
	public byte[] getOCSPSCT() {
		return OCSPSCT;
	}

	/**
	 * @return the tLSSCT
	 */
	public byte[] getTLSSCT() {
		return TLSSCT;
	}

	/**
	 * @return the x509SCT
	 */
	public byte[] getX509SCT() {
		return X509SCT;
	}

	/**
	 * @return the oCSPTime
	 */
	public long getOCSPTime() {
		return OCSPTime;
	}

	/**
	 * @return the tLSTime
	 */
	public long getTLSTime() {
		return TLSTime;
	}

	/**
	 * @return the x509Time
	 */
	public long getX509Time() {
		return X509Time;
	}

	/**
	 * @return the handshakeTime
	 */
	public long getHandshakeTime() {
		return handshakeTime;
	}

	/**
	 * @return the amountBytesSent
	 */
	public long getAmountBytesSent() {
		return amountBytesSent;
	}

	/**
	 * @return the amountBytesReceived
	 */
	public long getAmountBytesReceived() {
		return amountBytesReceived;
	}

	/**
	 * @return the certificate
	 */
	public X509Certificate getCertificate() {
		return certificate;
	}
	
	public boolean hasSCT()
	{
		return this.usedOCSPStapling || this.usedTLSExtension || this.usedX509v3Extension;
	}
	
	public String getIpAddress() {
		return ipAddress;
	}

	private boolean wait = true;
	
	private boolean usedTLSExtension = false;
	private boolean usedX509v3Extension = false;
	
	private byte[] OCSPSCT = new byte[]{};
	private byte[] TLSSCT = new byte[]{};
	private byte[] X509SCT = new byte[]{};
	
	private long OCSPTime = 0;
	private long TLSTime = 0;
	private long X509Time = 0;
	private long handshakeTime = 0;
	
	private long amountBytesSent = 0;
	private long amountBytesReceived = 0;
	
	private String ipAddress;
	
	private boolean fetchSCT;
	

	private X509Certificate certificate;
	
	public TLSClient(String host)
	{
		this.host = host;
		this.fetchSCT = true;
	}
	
	public TLSClient(String host, boolean fetchSCT)
	{
		this.host = host;
		this.fetchSCT = fetchSCT;
	}
	
	public void connect() throws IOException
	{
		Socket socket = new Socket(host, 443);
		this.ipAddress = socket.getInetAddress().getHostAddress();
		socket.setSoTimeout(10000);
		long starttime = System.nanoTime();
		CountingOutputStream cos = new CountingOutputStream(socket.getOutputStream());
		CountingInputStream cis = new CountingInputStream(socket.getInputStream());
		TlsClientProtocol tls = new TlsClientProtocol(cis, cos, new SecureRandom()) {
			@Override
			public void connect(TlsClient client) throws IOException {
				allowCertificateStatus = true;
				super.connect(client);
			}
			
			protected void sendClientHelloMessage() throws IOException{
				super.sendClientHelloMessage();
			}

			@Override
			protected void receiveServerHelloMessage(ByteArrayInputStream buf) throws IOException {
				super.receiveServerHelloMessage(buf);
				if(TLSClient.this.fetchSCT)
				{
					byte[] sct = (byte[]) serverExtensions.get(ExtensionType.signed_certificate_timestamp);
					if (sct != null) {
						TLSTime = System.nanoTime() - starttime;
						TLSSCT = sct;
						usedTLSExtension = true;
					}
				}
			}
			
			protected void handleHandshakeMessage(short arg0, byte[] arg1) throws IOException
			{
				super.handleHandshakeMessage(arg0, arg1);
				
			}

			@Override
			protected void completeHandshake() throws IOException {
				if(TLSClient.this.fetchSCT)
				{
					if (certificateStatus != null) {
						BasicOCSPResponse response = BasicOCSPResponse.getInstance(certificateStatus.getOCSPResponse().getResponseBytes().getResponse().getOctets());
						for (ASN1Encodable r : response.getTbsResponseData().getResponses()) {
							Extension extension = getExtension(SingleResponse.getInstance(r).getSingleExtensions(), OCSP_SIGNED_CERTIFICATE_TIMESTAMP);
							if (extension != null) {
								OCSPTime = System.nanoTime() - starttime;
								OCSPSCT = ASN1OctetString.getInstance(extension.getExtnValue().getOctets()).getOctets();
								usedOCSPStapling = true;
							}
						}
					}
				}
				amountBytesReceived = cis.getByteCount();
				amountBytesSent = cos.getByteCount();
				super.completeHandshake();
				handshakeTime = System.nanoTime() - starttime;
			}
		};

		TlsClient client = new DefaultTlsClient() {
			@Override
			public Hashtable getClientExtensions() throws IOException {
				CertificateStatusRequest statusRequest = new CertificateStatusRequest(CertificateStatusType.ocsp, new OCSPStatusRequest(null, null));
				@SuppressWarnings("unchecked")
				Hashtable<Integer, byte[]> extensions = TlsExtensionsUtils.ensureExtensionsInitialised(super.getClientExtensions());
				extensions.put(ExtensionType.signed_certificate_timestamp, new byte[0]);
				extensions.put(ExtensionType.status_request, TlsExtensionsUtils.createStatusRequestExtension(statusRequest));
				return extensions;
			}

			@Override
			public TlsAuthentication getAuthentication() throws IOException {
				return new ServerOnlyTlsAuthentication() {
					@Override
					public void notifyServerCertificate(Certificate certificate) throws IOException {
						if(TLSClient.this.fetchSCT)
						{
							Extension extension = getExtension(certificate.getCertificateAt(0).getTBSCertificate().getExtensions(), X509_SIGNED_CERTIFICATE_TIMESTAMP);
							if (extension != null) {
								X509Time = System.nanoTime() - starttime;
								X509SCT = ASN1OctetString.getInstance(extension.getExtnValue().getOctets()).getOctets();
								usedX509v3Extension = true;
							}
						}
						byte[] encoded = DatatypeConverter.parseBase64Binary(DatatypeConverter.printBase64Binary(certificate.getCertificateAt(0).getEncoded()));
						
						try {
							
							java.security.cert.X509Certificate jcert = (X509Certificate)CertificateFactory.getInstance("X.509").generateCertificate(new ByteArrayInputStream(encoded));
							TLSClient.this.certificate = (X509Certificate)jcert;

						} catch (CertificateException e) {
							// TODO Auto-generated catch block
							throw new IllegalArgumentException("Unknown certificate");
						}
					}
				};
			}
		};
		try {
			tls.connect(client);
		} finally {
			tls.close();
		}
	}

}
